#!/usr/bin/perl

#
# Created by Wim Lewis <wiml@hhhh.org>
# Copyright (c) 2003
#

use DB_File;
use strict;
use File::Find;
use MP3::Info;

use Data::Dumper;

our($index_fn, $tracks_fn, @music, $homedir, %dbkeys);

do 'rioserve.common' or die;

my ($IX, $DB, %btree, %tracks);

my $TRACK_INDEX = new DB_File::BTREEINFO;
$TRACK_INDEX->{'flags'} = R_DUP;

$IX = tie %btree, 'DB_File', $index_fn, O_RDWR|O_CREAT, 0666, $TRACK_INDEX
    or die "Cannot open $index_fn: $!\n";
$DB = tie %tracks, 'DB_File', $tracks_fn, O_RDWR|O_CREAT, 0666, $DB_HASH
    or die "Cannot open $tracks_fn: $!\n";

my @indexkeys = qw( artist source title genre );
my %vorbistagmap = qw(ARTIST artist ALBUM source TITLE title GENRE genre TRACKNUMBER track);
my %mp3tagmap = qw(artist artist album source title title genre genre bitrate bitrate secs duration tracknum track);

my %inverse;
my $nextid;
print "Building reverse index of tracks ...\n";
my $trackid = '';
my $trackloc = '';
my $ok = $DB->seq($trackid, $trackloc, R_FIRST);
my $count = 0;
while ($ok == 0) {
    unless ($trackid eq '__NEXT__') {
	my %trackinfo = split("\000", $trackloc);
	$inverse{$trackinfo{'@'}} = $trackid;
    }
    $ok = $DB->seq($trackid, $trackloc, R_NEXT);
    $count ++;
}
$nextid = $tracks{'__NEXT__'};
$nextid = 0x4001 unless $nextid;
print "(Scanned $count entries. ";
$count = scalar(keys %inverse);
print "Index contains $count tracks. Next track id is $nextid.)\n";
#exit 0;

print "Scanning directory for new files ...\n";
find(\&process, @music);
sub process {
    my $fn = $File::Find::name;
    my $filesize;

    return unless $fn =~ /\.(?:mp3|ogg|wav|flac)$/i;
    return if exists $inverse{$fn};

    return if -d $fn;
    $filesize = -s _;
    return if $filesize < 50000;

    my %info = ();
    my ($codec, $bitrate, $duration);

    my $mp3info = MP3::Info->new($fn);
    if ($mp3info) {
	%info = (
		 artist   => $mp3info->artist,
		 source   => $mp3info->album,
		 title    => $mp3info->title,
		 genre    => $mp3info->genre,
		 codec    => 'mp3',
		 bitrate  => $mp3info->bitrate,
		 duration => $mp3info->secs,
		 track    => $mp3info->tracknum
		 );
    } elsif ($fn =~ /\.ogg$/i) {
	$info{'codec'} = 'ogg';
	open (VORBINFO, '-|', "vorbiscomment", "-l", $fn) || die;
	while (<VORBINFO>) {
	    chomp;
	    my ($tag, $value) = split('=', $_, 2);
	    my $mapped = $vorbistagmap{uc($tag)};
	    if ($mapped) {
		$info{$mapped} = $value;
	    }
	}
	close (VORBINFO) || warn;
    } elsif($fn =~ /\.flac$/i) {
        $info{'codec'} = 'flac';
        &metaflac(\%info, $fn);
    }
	
    unless (%info) {
	print "? $fn\n";
	return;
    }

    $info{'filesize'} = $filesize;
    $info{'path'} = $fn;
    $trackid = sprintf("%04x", ++ $nextid);
    
    my $indexed = 0;
    my %trackinfo;

    foreach my $indicus (keys %info) {
	next unless $info{$indicus} and exists($dbkeys{$indicus});
	my $value = $info{$indicus};
	my $indexable = scalar(grep {$_ eq $indicus} @indexkeys);

	if ($indexable) {
	    # Get rid of any non-printable characters
	    $value =~ tr/\000-\037/ /;
	    # Remove leading and trailing spaces
	    $value =~ s/^\s+//;
	    $value =~ s/\s+$//;
	    # Collapse any strings of spaces
	    $value =~ s/\s\s+/ /g;
	    # Fix up some genre inconsistencies
	    if ($indicus eq 'genre') {
		$value = ucfirst($value);
		$value =~ s/sound\s*track/Soundtrack/i;
	    }
	}

	# Clean up track numbers
	if ($indicus eq 'track') {
	    $value =~ s-^(\d+)/.*$-$1-;
	    $value = 0 + $value;
	    next if $value == 0;
	}
	# Clean up other numbers 
	if ($indicus eq 'duration' or $indicus eq 'bitrate') {
	    $value = int(0.8 + $value);
	    next if $value == 0;
	}

	$trackinfo{$dbkeys{$indicus}} = $value;

	if ($indexable) {
	    # Turn any colons into semicolons to avoid confusion in rio protocol
	    $value =~ tr/:/;/;
	    # Move particles to the end
	    $value =~ s/^(a)\s+(.+)$/$2, $1/i;
	    $value =~ s/^(the)\s+(.+)$/$2, $1/i;
	    # Store it in the database
	    $IX->put($dbkeys{$indicus} . $value, $trackid);
	    $indexed = 1;
	}
    }
	
    if ($indexed) {
	print "+ $fn\n";
	$tracks{$trackid} = join("\000", %trackinfo);
    }
}

$tracks{'__NEXT__'} = $nextid;


undef $IX;
undef $DB;
untie %btree;
untie %tracks;


sub metaflac {
    my ($info, $fn) = @_;
    my ($blocktype, %streaminfo, %vorbiscomment, $k);
    
    open(FLACINFO, '-|', 'metaflac', '--list', 
	 '--block-type', 'STREAMINFO,VORBIS_COMMENT', $fn) || die;
    while(<FLACINFO>) {
	chomp;
	if (/^METADATA block /) {
	    undef $blocktype;
	    next;
	}
        print "($_)\n";
	if (/^\s+type\: \d+ \(([^)]*)\)/) {
            $blocktype = $1;
	    next;
	}
        if ($blocktype eq 'STREAMINFO' && /^\s+([^:]+): (.*$)/) {
	    $streaminfo{$1} = $2;
	    next;
	}
	if ($blocktype eq 'VORBIS_COMMENT' &&
	    /^\s+comment\[\d+\]\:\s+([^=]*)\=(.*$)/) {
	    $vorbiscomment{uc($1)} = $2;
	    next;
	}
    }
    close(FLACINFO) || warn;
    warn "No STREAMINFO block in $fn ?\n" unless %streaminfo;
	    
    if (defined($streaminfo{'sample_rate'}) &&
	defined($streaminfo{'total samples'})) {
        my($bits, $duration, $bitrate);

	$bits = 8 * -s $fn;
	$duration = $streaminfo{'total samples'} / $streaminfo{'sample_rate'};
	$info->{'duration'} = int( $duration );
	$info->{'bitrate'} = int( $bits / $duration / 1000.0 )
    }

    foreach $k (keys %vorbiscomment) {
	if(defined($vorbistagmap{$k})) {
	    $info->{$vorbistagmap{$k}} = $vorbiscomment{$k};
	}
    }
}

