#!/usr/bin/perl -w

#
# Inetd-based UDP server for auto-configuration of Dell Audio Radio.
# Based on the standalone version written by Jeff Mock.
#
# The Rio player makes a couple of requests on UDP port 21075 when
# it starts up, and none otehrwise. This script sends replies which
# point the client box to the appropriate servers.
#
# Typical line in inetd.conf:
#   21075 dgram udp wait rioserve /home/rioserve/ssdp-inetd.pl
# (assuming you have a user named rioserve; the UID under which this
# program runs is unimportant.)
#

use strict;
use POSIX;
use Socket;
use Sys::Syslog;

# Configuration variables.
# $mserve_ip must be a dotted-quad unless you modify the client
# NFS image to include /etc/resolv.conf.
#
# my $mserve_ip   = "192.168.10.1"; 
my $mserve_ip = "10.1.0.22";   # NFS and HTTP host
my $mserve_port = "81";       # HTTP port number

Sys::Syslog::setlogsock 'unix';  # default of 'inet' fails silently on obsd!
openlog('ssdp-rioserve', 'pid', 'user');
syslog('info', '%s started', $0);

#
# The box makes two different requests.  One comes from the kernel
# during initial booting, the second comes from the player application
# after the second boot when the player starts.
#
# The respones are different for Linux.  If there is a port number
# on the first "linux" request then the client box will use that port
# for portmapper lookups, which is generally bad when talking to 
# another linux box.
#
# The second "player" response includes a port number that indicates
# the port number to use for HTTP music related requests.  I use
# port 81 and setup a virtual server in Apache to respond to music
# requests, but you may want to do this differently.
#
my $player_request = "^upnp:uuid:1D274DB0-F053-11d3-BF72-0050DA\\w{6}";
my $linux_request  = "^upnp:uuid:1D274DB1-F053-11d3-BF72-0050DA\\w{6}";

my (
    $him,          # peer making UDP request
    $datagram,     # Packet from client
);

# Set a timer to keep the daemon from living forever.
# Recent perls (5.8.x and later) use a deferred signal handling mechanism,
# so it should be safe to call syslog from within the handler.
$SIG{'ALRM'} = sub { 
    syslog('info', '%s bored, exiting now', $0);
    exit 0;
};
alarm 30;

#
# Wait for requests and respond if appropriate
#
while ($him = recv(STDIN, $datagram, 256, 0)) {
    my ($port, $iaddr) = sockaddr_in($him);
    my $peer = inet_ntoa($iaddr);
    $datagram =~ s/\n//g;
    if ($datagram =~ $linux_request) {
	syslog('info', 'Linux request from %s.', $peer);
        send(STDIN, "http://$mserve_ip/descriptor.xml\n", 0, $him)
		or syslog('notice', "can't send response: %m");
    } elsif ($datagram =~ $player_request) {
	syslog('info', 'Player request from %s.', $peer);
        send(STDIN, "http://$mserve_ip:$mserve_port/descriptor.xml\n", 0, $him)
		or syslog('notice', "can't send response: %m"); 
    } else {
	syslog('notice|debug', 'Unhandled request from %s:%d', $peer, $iaddr);
    }
    syslog('info', 'done for now');
} 

syslog('info', '%s exiting: %m', $0);

exit 0;
